/**
*	WARNING: Do Not Remove This Section
*
*	$RCSfile: bitset.c,v $
*	$Revision: 1.6 $
*	$Date: 2006/10/11 15:49:54 $
*	$Author: zapreevis $
*	
*	MRMC is a model checker for discrete-time and continuous-time Markov reward models.
*	It supports reward extensions of PCTL and CSL (PRCTL and CSRL), and allows for the
*	automated verification of properties concerning long-run and instantaneous rewards
*	as well as cumulative rewards.
*	
*	Copyright (C) The University of Twente, 2004-2006.
*	Authors: Maneesh Khattri, Ivan Zapreev
*
*	This program is free software; you can redistribute it and/or
*	modify it under the terms of the GNU General Public License
*	as published by the Free Software Foundation; either version 2
*	of the License, or (at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
*	Our contacts:
*		Formal Methods and Tools Group, University of Twente,
*		P.O. Box 217, 7500 AE Enschede, The Netherlands,
*		Phone: +31 53 4893767, Fax: +31 53 4893247,
*		E-mail: mrmc@cs.utwente.nl
*
*	Source description: Manage a Bitset - For the satisfaction relation in MCC.
*	Uses: Definition of bitset - bitset.h
*/

#include "bitset.h"

//This macro comutes the number of bytes structure
//needed in order to hold the LENGTH bits
#define NUMBER_OF_BLOCKS(LENGTH) ( ( int ) ceil( ( ( double ) ( LENGTH ) ) / ( ( double ) BITSET_BLOCK_SIZE ) ) )
//First we compute the number of byte structures
//required to hols IDX+1 bits and then we subtract 1
//NOTE: 1. This "+1" is because indexing of states
//         and bitset elements start from 0
//      2. This "subtract 1" is because index in the
//         array of bytes starts from 0
#define IDX_TO_BLOCK_NUM(IDX) ( NUMBER_OF_BLOCKS( ( IDX ) + 1 ) - 1 )
//Simply do IDX mod BITSET_BLOCK_SIZE to compute
//the bit position within the byte structure
#define IDX_TO_IN_BLOCK_POSITION(IDX) ( ( IDX ) % BITSET_BLOCK_SIZE )

/**
* Get a new bitset.
* A bitset is composed of an array of BITSET_BLOCK_SIZE-bit unsigned integers.
* @param	: int n: indicates the size of the required bitset.
* @return	: bitset *: returns a pointer to a new bitset.
*/
bitset * get_new_bitset(int n)
{
	const int d = NUMBER_OF_BLOCKS(n);

	//Allocate the bitset
	bitset * newbitset = (bitset *)calloc(1,sizeof(bitset));
	
	//Allocate the bytes array
	if( n != 0 && d != 0 ){
		newbitset->bytesp = (bytes *)calloc(d,sizeof(bitset));
	}
	
	//Init the other bitset fields
	newbitset->n = n;
	newbitset->blocks_num = d;

	return newbitset;
}

/*****************************************************************************
name		: or
role		: obtain the disjunction of two bitsets.
@param		: bitset *a: One of the operands to the disjunction operation.
@param		: bitset *b: The second operand to the disjuction operation.
@return	: bitset *: returns a pointer to the result of disjunction.
remark		: The size of the operands must be the same else a NULL pointer is
		  returned as the result.
******************************************************************************/
bitset * or(bitset *a, bitset *b)
{
	int i, n = a->n;
	bitset *result = NULL;
	if( n == b->n ){
		result = get_new_bitset(n);
		for( i=0; i < a->blocks_num ; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits | b->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct a disjunction of two bitsets with different lenghts.\n");
	}
	return result;
}

/*****************************************************************************
name		: or_result
role       	: obtain the disjunction of two bitsets and put result in arg2.
@param     	: bitset *a: One of the operands to the disjunction operation.
@param		: bitset *b: The second operand to the disjunction operation.
******************************************************************************/
void or_result(bitset *a, bitset *result)
{
	int i;
	if( a->n == result->n ){
		for( i=0; i < a->blocks_num; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits | result->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct a disjunction of two bitsets with different lenghts.\n");
	}
}


/*****************************************************************************
name		: copy_bitset
role       	: This method just copies elements from one bitset into enother.
		: The two bitsets must have the same size.
@param     	: bitset *from_bitset: the bitset to be copied .
@param		: bitset *to_bitset: the bitset to copy to.
******************************************************************************/
void copy_bitset(bitset *from_bitset, bitset *to_bitset)
{
	int i;
	if( from_bitset->n == to_bitset->n ){
		for( i=0; i < from_bitset->blocks_num; i++ ){
			to_bitset->bytesp[i].bits = from_bitset->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to copy bitsets with different lenghts.\n");
	}
}

/*****************************************************************************
name		: xor
role       	: obtain the exclusive-or of two bitsets.
@param     	: bitset *a: One of the operands to the exclusive-or operation.
@param		: bitset *b: The second operand to the exclusive-or operation.
@return		: bitset *: returns a pointer to the result of exclusive-or
remark 	        : The size of the operands must be the same else a NULL pointer is
		    returned as the result.
******************************************************************************/
bitset * xor(const bitset *a, const bitset *b)
{
	int i, n=a->n;
	bitset *result = NULL;
	if( n == b->n ){
		result = get_new_bitset(n);
		for( i=0; i < a->blocks_num; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits ^ b->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct an exclusive disjunction of two bitsets with different lenghts.\n");
	}

	return result;
}

/*****************************************************************************
name		: xor_result
role       	: obtain the xor of two bitsets and put result in arg2.
@param     	: bitset *a: One of the operands to the x-or operation.
@param		: bitset *b: The second operand to the x-or operation.
******************************************************************************/
void xor_result(bitset *a, bitset *result)
{
	int i;
	if( a->n == result->n ){
		for( i = 0; i < a->blocks_num; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits ^ result->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct an exclusive disjunction of two bitsets with different lenghts.\n");
	}
}

/*****************************************************************************
name		: and
role       	: obtain the conjunction of two bitsets.
@param     	: bitset *a: One of the operands to the conjunction operation.
@param		: bitset *b: The second operand to the conjunction operation.
@return		: bitset *: returns a pointer to the result of conjunction.
remark 	        : The size of the operands must be the same else a NULL pointer is
		    returned as the result.
******************************************************************************/
bitset * and(const bitset *a, const bitset *b)
{
	int i, n=a->n;
	bitset *result = NULL;
	if( n == b->n ){
		result = get_new_bitset(n);
		for( i = 0; i < a->blocks_num; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits & b->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct a conjunction of two bitsets with different lenghts.\n");
	}

	return result;
}

/*****************************************************************************
name		: and_result
role       	: obtain the conjunction of two bitsets and put result in arg2.
@param     	: bitset *a: One of the operands to the conjunction operation.
@param		: bitset *b: The second operand to the conjunction operation.
******************************************************************************/
void and_result(bitset *a, bitset *result)
{
	int i;
	if( a->n == result->n ){
		for( i = 0; i < a->blocks_num; i++ ){
			result->bytesp[i].bits = a->bytesp[i].bits & result->bytesp[i].bits;
		}
	}else{
		printf("WARNING: Trying to construct a conjunction of two bitsets with different lenghts.\n");
	}
}

/*****************************************************************************
name		: not
role       	: obtain the complement of a bitset.
@param     	: bitset *a: The operand to the complement operation.
@return		: bitset *: returns a pointer to the result of complement.
remark 	        :
******************************************************************************/
bitset * not(const bitset *a)
{
	int i, n=a->n, d;
	bitset *result = NULL;
	bitset *dummy = get_new_bitset(n);
	
	fill_bitset_one(dummy);
	
	result = xor(a, dummy);
	
	free_bitset(dummy);
	return result;
}

/*****************************************************************************
name		: not_result
role       	: obtain the conjunction of two bitsets and put result in arg2.
@param     	: bitset *result: One of the operands to the complement operation.
@return	: the result pointer, for convenience
******************************************************************************/
bitset *not_result(bitset *result)
{
	int i, n=result->n, d;
	bitset *dummy = NULL;
	dummy = get_new_bitset(n); fill_bitset_one(dummy);
	xor_result(dummy, result);
	free_bitset(dummy);
	return result;
}

/*****************************************************************************
name		: get_bit_val
role       	: obtain the value of a specified bit of a bitset.
@param     	: bitset *a: One of the bits of this bitset is of interest.
@param		: int i: The position of the bit of interest in the bitset.
@param		: bit * (result by reference): The bit of interest is returned
		    	in this parameter.
@return		: int: indicates the success/failure of the get operation.
			 -1: invalid i, 1: success
remark 	        :
******************************************************************************/
int get_bit_val(const bitset *a, const int i, bit *result)
{
	int d, r;
	unsigned long int val;

	if( i < 0 || i >= a->n ){
		return -1;
	}

	d = IDX_TO_BLOCK_NUM(i);
	r = IDX_TO_IN_BLOCK_POSITION(i);

	val = 1<<r;
	val = (a->bytesp[d].bits&val)>>r;
	result->b = val;

	return 1;
}

/*****************************************************************************
name		: set_bit_val
role       	: set the value of a specified bit of a bitset.
@param     	: bitset *a: One of the bits of this bitset is to be changed.
@param		: int i: The position of the bit of interest in the bitset.
@param		: bit * val: The new value for the bit of interest.
@return		: int: indicates the success/failure of the get operation.
			 -1: invalid i, 1: success
remark 	        :
******************************************************************************/
void set_bit_val(bitset *a, int i, const bit *val)
{
	int d;
	unsigned long int value=1;

	if(i<0||i>=a->n)
	{
		printf("ERROR: trying to set a value to bit number %d, which is out of bound [0,%d]\n", i , a->n-1);
		exit(1);
	}

	d = IDX_TO_BLOCK_NUM(i);

	value <<= IDX_TO_IN_BLOCK_POSITION(i);

	if(val->b){
		a->bytesp[d].bits|=value;
	}else{
		a->bytesp[d].bits&=(~value);
	}
}

/*****************************************************************************
name		: print_bitset_states
role       	:print the position of the bits in the given bitset whose value
		 is 1.
@param     	: bitset *a: The bitset to be printed.
@return		: 
remark 	        :
******************************************************************************/
void print_bitset_states(bitset *a)
{
	int i;
	bit getb;
	if( a )
	{
		printf("states = {");
		for(i=0;i<a->n;i++)
		{
			get_bit_val(a,i,&getb);
			if(getb.b)
				printf("%d, ", i+1);
		}
		printf("}\n");
	}else{
		printf("WARNING: Trying to print a bitset for a NULL pointer.\n");
	}
}

/*****************************************************************************
name		: free_bitset
role       	: free the given bitset.
@param     	: bitset *a: The bitset to be freed.
@return		: 
remark 	        :
******************************************************************************/
void free_bitset(bitset *a)
{
	if( a )
	{
		if(a->bytesp) free(a->bytesp);
		free(a);
	}
}

/*****************************************************************************
name		: fill_bitset_one
role       	: fills the given bitset with one.
@param     	: bitset *a: The bitset to be filled.
@return		: bitset *: returns a pointer to the result of the filling
		  operation.
remark 	        :
******************************************************************************/
void fill_bitset_one(bitset *a)
{
	int i;
	unsigned long int val = pow(2,BITSET_BLOCK_SIZE)-1;	//TODO: CHECK IT!

	for( i=0; i < a->blocks_num; i++ ){
		a->bytesp[i].bits = a->bytesp[i].bits | val;
	}
}

/*****************************************************************************
name		: is_bitset_zero
role       	: checks if the bitset contains only zeros.
@param     	: bitset *a: The bitset to be checked.
@return		: 0: bitset contains a non-zero, 1: bitset contains only zeros.
******************************************************************************/
int is_bitset_zero(bitset *a)
{
	int i, flag=1;
	for( i=0; i < a->blocks_num; i++ ){
		if( a->bytesp[i].bits !=0 ){
			flag=0;
			break;
		}
	}
	return flag;
}

/**
* Get the Index of the next non-zero element.
* @param	: bitset *a: The bitset to be checked.
* @param	: int idx: The Present index.
*		NOTE: If set to -1 then it is the first time call
*		for this method on the given bitset i.e. we have
*		to start searching from the very beginning
* @return	: next index.
*/
int get_idx_next_non_zero(const bitset *a, const int idx)
{
	//If idx==-1 then we set it to 0 to make us searching
	//from the very first byte of the bitset
	const int p = IDX_TO_BLOCK_NUM(idx == -1 ? 0 : idx);
	//If idx==-1 then r will be -1 which will make us searching
	//from the very beginning of the byte
	const int r = IDX_TO_IN_BLOCK_POSITION(idx);
	//This is the number of bytes in the bitset
	const int NUMBER_OF_BYTES = a->blocks_num;
	//This is the index of the last byte in the bitset
	const int LAST_BLOCK_INDEX = NUMBER_OF_BYTES - 1;
	//This is the last valid bit index in the last byte of the bitset
	const int LAST_BIT_INDEX = IDX_TO_IN_BLOCK_POSITION( a->n - 1);

	int i, j, r_val;
	BOOL found = FALSE;
	unsigned long int val = 1;
	unsigned long int bits;
	
	//Set the starting bit for search in the byte number p
	j = r + 1;
	//Do search for the other non zerror elements
	for( i = p; i < NUMBER_OF_BYTES; i++ ){
		bits = a->bytesp[i].bits >> j;
		//Check if the current block is not empty
		if( bits ){
			//Here we should stop iterations if: 
			//1. idx was pointing to the last element of this byte subset
			//2. the rest of the byte subset are just zerro bits
			//3. If we are in the last byte of the bitset and we
			//   have checked the last VALID bit of the bitsetset i.e. LAST_BIT_INDEX
			while( j < BITSET_BLOCK_SIZE && bits != 0 && ( i < LAST_BLOCK_INDEX || j <= LAST_BIT_INDEX )){
				if( ( bits & val ) != 0 ){
					found = TRUE;
					break;
				}
				bits = bits >> 1;
				j++;
			}
			if( found ){
				//If a nonzerror bit is found then break
				break;
			}
		}
		//In the next byte we start searching from the very first bit.
		j = 0;
	}
	if( found ){
		r_val = i * BITSET_BLOCK_SIZE + j;
	}else{
		r_val = -1;
	}
	return r_val;
}

/*****************************************************************************
name		: count_non_zero
role       	: count the number of non-zero elements in the given bitset.
@param     	: bitset *a: The bitset to be checked.
@return		: the count.
******************************************************************************/
int count_non_zero(const bitset *a)
{
	int i=-1, count=0;
	while( ( i = get_idx_next_non_zero(a, i) ) != -1 ) ++count;
	return count;
}

/**
* Count the number of valid bits in a bitset.
* put all the valid-bit ids in an int-array.
* the first element of the array is the count.
* the rest are the bit ids of valid (set) bits.
* @param	: bitset *toCount: the bitset in question
* @return	: the count of bits that are set and their ids.
*/
int * count_set(bitset *toCount)
{
	int size = toCount->n, i;
	bit getb;
	int *pValidSet = (int *) calloc (1, sizeof(int));
	pValidSet[0] = 0;
	for(i=0;i<size;i++)
	{
		get_bit_val(toCount, i, &getb);
		if(getb.b)
		{
			++pValidSet[0];
			pValidSet = (int *) realloc (pValidSet, (pValidSet[0]+1) * sizeof(int));
			pValidSet[pValidSet[0]] = i;
		}
	}
	return pValidSet;
}
